#!/usr/bin/python3.8
# -*- coding: utf-8 -*-

import json
from urllib.parse import urlencode
from exchanges_wrapper import __version__
import logging
import aiohttp
import hashlib
import hmac
import time
from errors import (
    RateLimitReached,
    BinanceError,
    WAFLimitViolated,
    IPAddressBanned,
    HTTPError,
    QueryCanceled,
)

logger = logging.getLogger('exch_srv_logger')


class HttpClient:
    def __init__(self, api_key, api_secret, endpoint, user_agent, proxy,
                 session, exchange='binance', sub_account=None):
        self.api_key = api_key
        self.api_secret = api_secret
        self.endpoint = endpoint
        self.rate_limit_reached = False
        if user_agent:
            self.user_agent = user_agent
        else:
            self.user_agent = f"exchange-wrapper, {__version__}"
        self.proxy = proxy
        self.session = session if session else aiohttp.ClientSession()
        self.exchange = exchange
        self.sub_account = sub_account

    def _generate_signature(self, data):
        return hmac.new(self.api_secret.encode("utf-8"), data.encode("utf-8"), hashlib.sha256).hexdigest()

    async def handle_errors(self, response):
        if response.status >= 500:
            logger.warning(f"An issue occurred on exchange's side: {response}")
            return {'success': False}
        if response.status == 429:
            self.rate_limit_reached = True if self.exchange == 'binance' else None
            raise RateLimitReached()
        # print(f"handle_errors.response: {response}")
        payload = await response.json()
        if payload and "code" in payload:
            # as defined here: https://github.com/binance/binance-spot-api-docs/blob/
            # master/errors.md#error-codes-for-binance-2019-09-25
            raise BinanceError(payload["msg"])
        if response.status >= 400:
            logger.error(f"handle_errors.response.status >= 400: {payload}")
            if response.status == 403:
                raise WAFLimitViolated()
            elif response.status == 418:
                raise IPAddressBanned()
            else:
                raise HTTPError(f"Malformed request: {payload}")
        # print(f"handle_errors.payload: {payload}")
        return payload

    async def send_api_call(
        self, path, method="GET", signed=False, send_api_key=True, **kwargs
    ):
        if self.rate_limit_reached:
            raise QueryCanceled(
                "Rate limit reached, to avoid an IP ban, this query has been automatically cancelled"
            )
        # return the JSON body of a call to Binance REST API
        # print(f"send_api_call.kwargs: {kwargs}")
        query_kwargs = {}
        content = str()
        ftx_post = self.exchange == 'ftx' and method == 'POST'
        _params = json.dumps(kwargs) if ftx_post else None
        url = f'{self.endpoint}{path}' if self.exchange == 'binance' else f'{self.endpoint}/{path}'
        if self.exchange == 'binance':
            query_kwargs = dict({"headers": {"User-Agent": self.user_agent}}, **kwargs,)
            if send_api_key:
                query_kwargs["headers"]["X-MBX-APIKEY"] = self.api_key
        elif self.exchange == 'ftx':
            # https://help.ftx.com/hc/en-us/articles/360052595091-2020-11-20-Ratelimit-Updates
            query_kwargs = {"headers": {"FTX-KEY": self.api_key}}
            content += urlencode(kwargs, safe='/')
            # print(f"send_api_call.content: {content}")
            if content and not ftx_post:
                url += f'?{content}'
            # print(f"send_api_call.url: {url}")
            if self.sub_account:
                query_kwargs["headers"]["FTX-SUBACCOUNT"] = self.sub_account
        if signed:
            if self.exchange == 'binance':
                location = "params" if "params" in kwargs else "data"
                query_kwargs[location]["timestamp"] = str(int(time.time() * 1000))
                if "params" in kwargs:
                    content += urlencode(kwargs["params"])
                if "data" in kwargs:
                    content += urlencode(kwargs["data"])
                query_kwargs[location]["signature"] = self._generate_signature(content)
                if self.proxy:
                    query_kwargs["proxy"] = self.proxy
            elif self.exchange == 'ftx':
                if ftx_post:
                    query_kwargs["headers"]["Content-Type"] = 'application/json'
                    query_kwargs.update({'data': _params})
                    content = f"{_params}"
                ts = int(time.time() * 1000)
                signature_payload = f'{ts}{method}/api/{path}'
                if content:
                    if ftx_post:
                        signature_payload += f'{content}'
                    else:
                        signature_payload += f'?{content}'
                # print(f"send_api_call.signature_payload: {signature_payload}")
                signature = self._generate_signature(signature_payload)
                query_kwargs["headers"]["FTX-SIGN"] = signature
                query_kwargs["headers"]["FTX-TS"] = str(ts)
        # logger.debug(f"send_api_call: method: {method}, url: {url}, **query_kwargs: {query_kwargs}")
        async with self.session.request(method, url, **query_kwargs) as response:
            # logger.debug(f"send_api_call.response: {response}")
            return await self.handle_errors(response)

    async def close_session(self):
        await self.session.close()



#!/usr/bin/python3.8
# -*- coding: utf-8 -*-

import asyncio
import functools
from collections import defaultdict

from errors import UnknownEventType


# based on: https://stackoverflow.com/a/2022629/10144963
class Handlers(list):
    async def __call__(self, *args, **kwargs):
        loop = asyncio.get_running_loop()
        for func in self:
            if asyncio.iscoroutinefunction(func):
                await func(*args, **kwargs)
                continue
            if kwargs:
                func = functools.partial(func, **kwargs)
            loop.run_in_executor(None, func, *args)

    def __repr__(self):
        return "Handlers(%s)" % list.__repr__(self)

# HANDLERS
# Example usage:

# from exchange-wrapper import events
#
# def my_order_update_listener(wrapped_event):
#    print(f"order for symbol {wrapped_event.symbol} updated!")
#
# events.order_update_handlers.append(my_order_update_listener)


class Events:
    def __init__(self):
        self.handlers = defaultdict(Handlers)
        self.registered_streams = defaultdict(set)

    def register_user_event(self, listener, event_type):
        self.handlers[event_type].append(listener)

    def unregister_user_event(self, event_type):
        self.handlers.pop(event_type)

    def register_event(self, listener, event_type, exchange='binance'):
        # print(f"register_event.listener: {listener}, event_type: {event_type}, exchange: {exchange}")
        self.registered_streams[exchange] |= {event_type}
        event_type = f"{event_type.split('@')[0].replace('/', '').lower()}@{event_type.split('@')[1]}"
        self.handlers[event_type].append(listener)
        # print(f"register_event.registered_streams: {self.registered_streams}")
        # print(f"register_event.handlers: {self.handlers}")

    def unregister(self, event_type, exchange='binance'):
        # print(f"unregister.event_type: {event_type}")
        self.registered_streams[exchange].discard(event_type)
        event_type = f"{event_type.split('@')[0].replace('/', '').lower()}@{event_type.split('@')[1]}"
        self.handlers.pop(event_type, None)
        # print(f"unregister.handlers: {self.handlers}")
        # print(f"unregister.registered_streams: {self.registered_streams}")

    def wrap_event(self, event_data):
        wrapper_by_type = {
            "outboundAccountPosition": OutboundAccountPositionWrapper,
            "balanceUpdate": BalanceUpdateWrapper,
            "executionReport": OrderUpdateWrapper,
            "listStatus": ListStatus,
            "aggTrade": AggregateTradeWrapper,
            "trade": TradeWrapper,
            "kline": KlineWrapper,
            "24hrMiniTicker": SymbolMiniTickerWrapper,
            "24hrTicker": SymbolTickerWrapper,
            "bookTicker": SymbolBookTickerWrapper,
            "depth5": PartialBookDepthWrapper,
            "depth10": PartialBookDepthWrapper,
            "depth20": PartialBookDepthWrapper,
            "depth": DiffDepthWrapper,
        }

        stream = event_data["stream"] if "stream" in event_data else False
        event_type = event_data["e"] if "e" in event_data else stream
        if "@" in event_type:  # lgtm [py/member-test-non-container]
            event_type = event_type.split("@")[1]
        if event_type.startswith("kline_"):
            event_type = "kline"
        if event_type not in wrapper_by_type:
            raise UnknownEventType()
        wrapper = wrapper_by_type[event_type]
        return wrapper(event_data, self.handlers[stream if stream else event_type])


class EventWrapper:
    def __init__(self, _event_data, handlers):
        self.handlers = handlers

    async def fire(self):
        if self.handlers:
            await self.handlers(self)


# MARKET EVENTS


class AggregateTradeWrapper(EventWrapper):
    def __init__(self, event_data, handlers):  # lgtm [py/similar-function]
        super().__init__(event_data, handlers)
        self.event_type = event_data["e"]
        self.event_time = event_data["E"]
        self.symbol = event_data["s"]
        self.aggregated_trade_id = event_data["a"]
        self.price = event_data["p"]
        self.quantity = event_data["q"]
        self.first_trade_id = event_data["f"]
        self.last_trade_id = event_data["l"]
        self.trade_time = event_data["T"]
        self.buyer_is_marker = event_data["m"]
        self.ignore = event_data["M"]


class TradeWrapper(EventWrapper):
    def __init__(self, event_data, handlers):  # lgtm [py/similar-function]
        super().__init__(event_data, handlers)
        self.event_type = event_data["e"]
        self.event_time = event_data["E"]
        self.symbol = event_data["s"]
        self.trade_id = event_data["t"]
        self.price = event_data["p"]
        self.quantity = event_data["q"]
        self.buyer_order_id = event_data["b"]
        self.seller_order_id = event_data["a"]
        self.trade_time = event_data["T"]
        self.buyer_is_marker = event_data["m"]
        self.ignore = event_data["M"]


class KlineWrapper(EventWrapper):
    def __init__(self, event_data, handlers):
        super().__init__(event_data, handlers)
        self.event_type = event_data["e"]
        self.event_time = event_data["E"]
        self.symbol = event_data["s"]
        kline = event_data["k"]
        self.kline_start_time = kline["t"]
        self.kline_close_time = kline["T"]
        self.kline_symbol = kline["s"]
        self.kline_interval = kline["i"]
        self.kline_first_trade_id = kline["f"]
        self.kline_last_trade_id = kline["L"]
        self.kline_open_price = kline["o"]
        self.kline_close_price = kline["c"]
        self.kline_high_price = kline["h"]
        self.kline_low_price = kline["l"]
        self.kline_base_asset_volume = kline["v"]
        self.kline_trades_number = kline["n"]
        self.kline_closed = kline["x"]
        self.kline_quote_asset_volume = kline["q"]
        self.kline_taker_buy_base_asset_volume = kline["V"]
        self.kline_taker_buy_quote_asset_volume = kline["Q"]
        self.kline_ignore = kline["B"]


class SymbolMiniTickerWrapper(EventWrapper):
    def __init__(self, event_data, handlers):  # lgtm [py/similar-function]
        super().__init__(event_data, handlers)
        self.event_type = event_data["e"]
        self.event_time = event_data["E"]
        self.symbol = event_data["s"]
        self.close_price = event_data["c"]
        self.open_price = event_data["o"]
        self.high_price = event_data["h"]
        self.low_price = event_data["l"]
        self.total_traded_base_asset_volume = event_data["v"]
        self.total_traded_quote_asset_volume = event_data["q"]


class SymbolTickerWrapper(EventWrapper):
    def __init__(self, event_data, handlers):
        super().__init__(event_data, handlers)
        self.event_type = event_data["e"]
        self.event_time = event_data["E"]
        self.symbol = event_data["s"]
        self.price_change = event_data["p"]
        self.price_change_percent = event_data["P"]
        self.weighted_average_price = event_data["w"]
        self.first_trade_before_window = event_data["x"]
        self.last_price = event_data["c"]
        self.last_quantity = event_data["Q"]
        self.best_bid_price = event_data["b"]
        self.best_bid_quantity = event_data["B"]
        self.best_ask_price = event_data["a"]
        self.best_ask_quantity = event_data["A"]
        self.open_price = event_data["o"]
        self.high_price = event_data["h"]
        self.low_price = event_data["l"]
        self.total_traded_base_asset_volume = event_data["v"]
        self.total_traded_quote_asset_volume = event_data["q"]
        self.statistics_open_time = event_data["O"]
        self.statistics_close_time = event_data["C"]
        self.first_trade_id = event_data["F"]
        self.last_trade_id = event_data["L"]
        self.total_trade_numbers = event_data["n"]


class SymbolBookTickerWrapper(EventWrapper):
    def __init__(self, event_data, handlers):
        super().__init__(event_data, handlers)
        self.order_book_updated = event_data["u"]
        self.symbol = event_data["s"]
        self.best_bid_price = event_data["b"]
        self.best_bid_quantity = event_data["B"]
        self.best_ask_price = event_data["a"]
        self.best_ask_quantity = event_data["A"]


class PartialBookDepthWrapper(EventWrapper):
    def __init__(self, event_data, handlers):
        super().__init__(event_data, handlers)
        self.last_update_id = event_data["lastUpdateId"]
        self.bids = event_data["bids"]
        self.asks = event_data["asks"]


class DiffDepthWrapper(EventWrapper):
    def __init__(self, event_data, handlers):
        super().__init__(event_data, handlers)
        self.event_type = event_data["e"]
        self.event_time = event_data["E"]
        self.symbol = event_data["s"]
        self.first_update_id = event_data["U"]
        self.final_update_id = event_data["u"]
        self.bids = event_data["b"]
        self.asks = event_data["a"]


# ACCOUNT UPDATE


class OutboundAccountPositionWrapper(EventWrapper):
    def __init__(self, event_data, handlers):
        super().__init__(event_data, handlers)
        self.event_time = event_data["E"]
        self.last_update = event_data["u"]
        self.balances = dict(
            map(lambda x: (x["a"], {"free": x["f"], "locked": x["l"]}), event_data["B"])
        )


# BALANCE UPDATE


class BalanceUpdateWrapper(EventWrapper):
    def __init__(self, event_data, handlers):
        super().__init__(event_data, handlers)
        self.event_time = event_data["E"]
        self.asset = event_data["a"]
        self.balance_delta = event_data["d"]
        self.clear_time = event_data["T"]


# ORDER UPDATE


class OrderUpdateWrapper(EventWrapper):
    def __init__(self, event_data, handlers):
        super().__init__(event_data, handlers)
        self.event_time = event_data["E"]
        self.symbol = event_data["s"]
        self.client_order_id = event_data["c"]
        self.side = event_data["S"]
        self.order_type = event_data["o"]
        self.time_in_force = event_data["f"]
        self.order_quantity = event_data["q"]
        self.order_price = event_data["p"]
        self.stop_price = event_data["P"]
        self.iceberg_quantity = event_data["F"]
        self.order_list_id = event_data["g"]
        self.original_client_id = event_data["C"]
        self.execution_type = event_data["x"]
        self.order_status = event_data["X"]
        self.order_reject_reason = event_data["r"]
        self.order_id = event_data["i"]
        self.last_executed_quantity = event_data["l"]
        self.cumulative_filled_quantity = event_data["z"]
        self.last_executed_price = event_data["L"]
        self.commission_amount = event_data["n"]
        self.commission_asset = event_data["N"]
        self.transaction_time = event_data["T"]
        self.trade_id = event_data["t"]
        self.ignore_a = event_data["I"]
        self.in_order_book = event_data["w"]
        self.is_maker_side = event_data["m"]
        self.ignore_b = event_data["M"]
        self.order_creation_time = event_data["O"]
        self.quote_asset_transacted = event_data["Z"]
        self.last_quote_asset_transacted = event_data["Y"]
        self.quote_order_quantity = event_data["Q"]


class ListStatus(EventWrapper):
    def __init__(self, event_data, handlers):
        super().__init__(event_data, handlers)
        self.event_time = event_data["E"]
        self.symbol = event_data["s"]
        self.order_list_id = event_data["g"]
        self.contingency_type = event_data["c"]
        self.list_status_type = event_data["l"]
        self.list_order_status = event_data["L"]
        self.list_reject_reason = event_data["r"]
        self.list_client_order_id = event_data["C"]
        # noinspection PyArgumentList
        self.orders = dict(
            map(lambda x: (x["s"], {"orderId": x["i"], "clientOrderId": x["c"]})),
            event_data["O"],
        )



"""
REST API and WebSocket asyncio wrapper with grpc powered multiplexer server for crypto exchanges (Binance, FTX, ).
Utilizes one connection for many accounts and trading pairs.
For SPOT market only
"""

__authors__ = ["Th0rgal", "Jerry Fedorenko"]
__license__ = "MIT"
__maintainer__ = "Jerry Fedorenko"
__contact__ = "https://github.com/DogsTailFarmer"
__email__ = "jerry.fedorenko@yahoo.com"
__version__ = "1.0b0"



#!/usr/bin/python3.8
# -*- coding: utf-8 -*-

from enum import Enum
from typing import Union
import decimal
import math
import asyncio
import logging
import random

from http_client import HttpClient
from errors import BinancePyError, RateLimitReached
from web_sockets import UserEventsDataStream, MarketEventsDataStream, FtxPrivateEventsDataStream
from definitions import OrderType
from events import Events

from ftx_parser import *

logger = logging.getLogger('exch_srv_logger')

STATUS_TIMEOUT = 5  # sec


def truncate(f, n):
    return math.floor(f * 10 ** n) / 10 ** n


class Client:
    def __init__(
        self,
        api_key=None,
        api_secret=None,
        *,
        endpoint="https://api.binance.com",
        user_agent=None,
        proxy=None,
        session=None,
        sub_account=None,
        endpoint_ws="wss://stream.binance.com:9443",
    ):
        if api_secret + api_secret == 1:
            raise ValueError(
                "You cannot only specify a non empty api_key or an api_secret."
            )
        # print(f"Client.api_key: {api_key}, api_secret: {api_secret}, endpoint: {endpoint}")
        self.api_key = api_key
        self.api_secret = api_secret
        self.sub_account = sub_account
        self.endpoint_ws = endpoint_ws
        if 'ftx' in endpoint:
            self.exchange = 'ftx'
        else:
            self.exchange = 'binance'
        self.http = HttpClient(
            api_key, api_secret, endpoint, user_agent, proxy, session,
            exchange=self.exchange, sub_account=self.sub_account,
        )
        self.user_agent = user_agent
        self.proxy = None
        self.loaded = False
        self.symbols = {}
        self.highest_precision = None
        self.rate_limits = None
        self.market_data_streams = []
        self.user_data_stream = None

    async def load(self):
        infos = await self.fetch_exchange_info()

        # load available symbols
        self.highest_precision = 8

        original_symbol_infos = infos["symbols"]
        for symbol_infos in original_symbol_infos:
            symbol = symbol_infos.pop("symbol")
            precision = symbol_infos["baseAssetPrecision"]
            if precision > self.highest_precision:
                self.highest_precision = precision
            symbol_infos["filters"] = dict(
                map(lambda x: (x.pop("filterType"), x), symbol_infos["filters"])
            )
            self.symbols[symbol] = symbol_infos

        # print(self.symbols)

        decimal.getcontext().prec = (
            self.highest_precision + 4
        )  # for operations and rounding

        # load rate limits
        self.rate_limits = infos["rateLimits"]

        self.loaded = True

    async def close(self):
        await self.http.close_session()

    @property
    def events(self):
        if not hasattr(self, "_events"):
            # noinspection PyAttributeOutsideInit
            self._events = Events()
        return self._events

    async def start_user_events_listener(self, endpoint=None):
        _endpoint = endpoint or "wss://stream.binance.com:9443"
        if self.exchange == 'binance':
            self.user_data_stream = UserEventsDataStream(self, _endpoint, self.user_agent)
            await self.user_data_stream.start()
        else:
            self.user_data_stream = FtxPrivateEventsDataStream(self,
                                                               _endpoint,
                                                               self.user_agent,
                                                               self.exchange,
                                                               self.sub_account
                                                               )
            await self.user_data_stream.start()

    async def stop_user_events_listener(self):
        # print(f"stop_user_events_listener()")
        if self.user_data_stream:
            await self.user_data_stream.stop()

    async def start_market_events_listener(self):
        _events = self.events.registered_streams
        logger.debug(f"start_market_events_listener._event: {_events}")
        start_list = []
        for _exchange in _events.keys():
            if _exchange == 'binance':
                _endpoint = "wss://stream.binance.com:9443"
                market_data_stream = MarketEventsDataStream(self, _endpoint, self.user_agent, _exchange)
                self.market_data_streams.append(market_data_stream)
                start = market_data_stream.start()
                start_list.append(start)
            else:
                _endpoint = self.endpoint_ws
                for channel in self.events.registered_streams.get(_exchange):
                    market_data_stream = MarketEventsDataStream(self, _endpoint, self.user_agent, _exchange, channel)
                    self.market_data_streams.append(market_data_stream)
                    start = market_data_stream.start()
                    start_list.append(start)
        logger.debug(f"start_market_events_listener.market_data_streams: {self.market_data_streams}")
        await asyncio.gather(*start_list, return_exceptions=False)

    async def stop_market_events_listener(self):
        # print(f"stop_market_events_listener()")
        stop_list = []
        for market_data_stream in self.market_data_streams:
            stop = market_data_stream.stop()
            stop_list.append(stop)
        await asyncio.gather(*stop_list, return_exceptions=False)

    def assert_symbol_exists(self, symbol):
        if self.loaded:
            if symbol not in self.symbols:
                raise BinancePyError(
                    f"Symbol {symbol} is not valid according to the loaded exchange infos."
                )

    def symbol_to_ftx(self, symbol) -> str:
        symbol_info = self.symbols.get(symbol)
        return f"{symbol_info.get('baseAsset')}/{symbol_info.get('quoteAsset')}"

    def refine_amount(self, symbol, amount: Union[str, decimal.Decimal], quote=False):
        if type(amount) == str:  # to save time for developers
            amount = decimal.Decimal(amount)
        if self.loaded:
            precision = self.symbols[symbol]["baseAssetPrecision"]
            lot_size_filter = self.symbols[symbol]["filters"]["LOT_SIZE"]
            step_size = decimal.Decimal(lot_size_filter["stepSize"])
            # noinspection PyStringFormat
            amount = (
                (f"%.{precision}f" % truncate(amount if quote else (amount - amount % step_size), precision))
                .rstrip("0")
                .rstrip(".")
            )
        return amount

    def refine_price(
        self, symbol, price: Union[str, decimal.Decimal]
    ) -> decimal.Decimal:
        if isinstance(price, str):  # to save time for developers
            price = decimal.Decimal(price)

        if self.loaded:
            precision = self.symbols[symbol]["baseAssetPrecision"]
            price_filter = self.symbols[symbol]["filters"]["PRICE_FILTER"]
            price = price - (price % decimal.Decimal(price_filter["tickSize"]))
            # noinspection PyStringFormat
            price = (
                (f"%.{precision}f" % truncate(price, precision))
                .rstrip("0")
                .rstrip(".")
            )
        return price

    def assert_symbol(self, symbol):
        if not symbol:
            raise ValueError("This query requires a symbol.")
        self.assert_symbol_exists(symbol)

    # keep support for hardcoded string but allow enums usage
    @staticmethod
    def enum_to_value(enum):
        if isinstance(enum, Enum):
            enum = enum.value
        return enum

    # GENERAL ENDPOINTS

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#test-connectivity
    async def ping(self):
        return await self.http.send_api_call("/api/v3/ping", send_api_key=False)

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#check-server-time
    async def fetch_server_time(self):
        return await self.http.send_api_call("/api/v3/time", send_api_key=False)

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#exchange-information
    async def fetch_exchange_info(self):
        binance_res = {}
        if self.exchange == 'binance':
            binance_res = await self.http.send_api_call(
                "/api/v3/exchangeInfo",
                send_api_key=False
            )
        elif self.exchange == 'ftx':
            res = await self.http.send_api_call(
                "markets",
                send_api_key=False
            )
            # Convert FTX to Binance Response
            if res and res.get('success'):
                binance_res = ftx_exchange_info(res.get('result'))
        # print(f"fetch_exchange_info.binance_res: {binance_res}")
        return binance_res

    # MARKET DATA ENDPOINTS

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#order-book
    async def fetch_order_book(self, symbol, limit=100):
        self.assert_symbol(symbol)
        valid_limits = []
        if self.exchange == 'binance':
            valid_limits = [5, 10, 20, 50, 100, 500, 1000, 5000]
        elif self.exchange == 'ftx':
            valid_limits = [5, 10, 20, 50, 100]
        binance_res = {}
        if limit in valid_limits:
            if self.exchange == 'binance':
                binance_res = await self.http.send_api_call(
                    "/api/v3/depth",
                    params={"symbol": symbol, "limit": limit},
                    send_api_key=False,
                )
            elif self.exchange == 'ftx':
                params = {'depth': limit}
                res = await self.http.send_api_call(
                    f"markets/{self.symbol_to_ftx(symbol)}/orderbook",
                    send_api_key=False,
                    **params,
                )
                # print(f"fetch_order_book.res: {res}")
                if res and res.get('success'):
                    binance_res = ftx_order_book(res.get('result'))
        else:
            raise ValueError(
                f"{limit} is not a valid limit. Valid limits: {valid_limits}"
            )
        # print(f"fetch_order_book.binance_res: {binance_res}")
        return binance_res

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#recent-trades-list
    async def fetch_recent_trades_list(self, symbol, limit=500):
        self.assert_symbol(symbol)
        if limit == 500:
            params = {"symbol": symbol}
        elif 0 < limit <= 1000:
            params = {"symbol": symbol, "limit": limit}
        else:
            raise ValueError(
                f"{limit} is not a valid limit. A valid limit should be > 0 and <= to 1000."
            )
        return await self.http.send_api_call(
            "/api/v3/trades", params=params, signed=False
        )

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#old-trade-lookup-market_data
    async def fetch_old_trades_list(self, symbol, from_id=None, limit=500):
        self.assert_symbol(symbol)
        if limit == 500:
            params = {"symbol": symbol}
        elif 0 < limit <= 1000:
            params = {"symbol": symbol, "limit": limit}
        else:
            raise ValueError(
                f"{limit} is not a valid limit. A valid limit should be > 0 and <= to 1000."
            )
        if from_id:
            params["fromId"] = from_id
        return await self.http.send_api_call(
            "/api/v3/historicalTrades", params=params, signed=False
        )

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#compressedaggregate-trades-list
    async def fetch_aggregate_trades_list(
        self, symbol, from_id=None, start_time=None, end_time=None, limit=500
    ):
        self.assert_symbol(symbol)
        if limit == 500:
            params = {"symbol": symbol}
        elif 0 < limit <= 1000:
            params = {"symbol": symbol, "limit": limit}
        else:
            raise ValueError(
                f"{limit} is not a valid limit. A valid limit should be > 0 and <= to 1000."
            )
        if from_id:
            params["fromId"] = from_id
        if start_time:
            params["startTime"] = start_time
        if end_time:
            params["endTime"] = end_time
        return await self.http.send_api_call(
            "/api/v3/aggTrades", params=params, signed=False
        )

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#klinecandlestick-data
    async def fetch_klines(self, symbol, interval, start_time=None, end_time=None, limit=500):
        self.assert_symbol(symbol)
        interval = self.enum_to_value(interval)
        if self.exchange == 'ftx':
            interval = ftx_interval(interval)
        if not interval:
            raise ValueError("This query requires correct interval value")
        binance_res = []
        if self.exchange == 'binance':
            if limit == 500:
                params = {"symbol": symbol, "interval": interval}
            elif 0 < limit <= 1000:
                params = {"symbol": symbol, "interval": interval, "limit": limit}
            else:
                raise ValueError(
                    f"{limit} is not a valid limit. A valid limit should be > 0 and <= to 1000."
                )
            if start_time:
                params["startTime"] = start_time
            if end_time:
                params["endTime"] = end_time
            binance_res = await self.http.send_api_call(
                "/api/v3/klines", params=params, signed=False
            )
        elif self.exchange == 'ftx':
            end_time = int(time.time())
            start_time = end_time - interval * limit - 1
            params = {
                'resolution': interval,
                'start_time': start_time,
                'end_time': end_time}
            res = await self.http.send_api_call(
                f"markets/{self.symbol_to_ftx(symbol)}/candles",
                send_api_key=False,
                **params,
            )
            # print(f"fetch_klines.res: {res}")
            if res and res.get('success'):
                # print(f"fetch_klines.len: {len(res.get('result'))}")
                # print(f"fetch_klines.res: {res.get('result')}")
                binance_res = ftx_klines(res.get('result'), interval)
        # print(f"fetch_klines.binance_res: {binance_res}")
        return binance_res

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#current-average-price
    async def fetch_average_price(self, symbol):
        self.assert_symbol(symbol)
        return await self.http.send_api_call(
            "/api/v3/avgPrice",
            params={"symbol": symbol},
            signed=False,
            send_api_key=False,
        )

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#24hr-ticker-price-change-statistics
    async def fetch_ticker_price_change_statistics(self, symbol=None):
        if symbol:
            self.assert_symbol_exists(symbol)
            binance_res = {}
        else:
            binance_res = []
        if self.exchange == 'binance':
            binance_res = await self.http.send_api_call(
                "/api/v3/ticker/24hr",
                params={"symbol": symbol} if symbol else {},
                signed=False,
                send_api_key=False,
            )
        elif self.exchange == 'ftx' and symbol:
            resolution = 60 * 60 * 24
            end_time = int(time.time())
            start_time = end_time - resolution
            params = {
                'resolution': resolution,
                'start_time': start_time,
                'end_time': end_time}
            res = await self.http.send_api_call(
                f"markets/{self.symbol_to_ftx(symbol)}/candles",
                send_api_key=False,
                **params,
            )
            # print(f"fetch_ticker_price_change_statistics.res: {res}")
            if res and res.get('success'):
                binance_res = ftx_ticker_price_change_statistics(res.get('result'), symbol, end_time)
        # print(f"fetch_ticker_price_change_statistics.binance_res: {binance_res}")
        return binance_res

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#symbol-price-ticker
    async def fetch_symbol_price_ticker(self, symbol=None):
        # print(f"fetch_symbol_price_ticker.symbol: {symbol}")
        if symbol:
            self.assert_symbol_exists(symbol)
            binance_res = {}
        else:
            binance_res = []
        if self.exchange == 'binance':
            binance_res = await self.http.send_api_call(
                "/api/v3/ticker/price",
                params={"symbol": symbol} if symbol else {},
                signed=False,
                send_api_key=False,
            )
        elif self.exchange == 'ftx':
            res = await self.http.send_api_call(
                f"markets/{self.symbol_to_ftx(symbol)}" if symbol else "markets",
                send_api_key=False,
            )
            # print(f"fetch_symbol_price_ticker.res: {res}")
            if res and res.get('success'):
                binance_res = ftx_symbol_price_ticker(res.get('result'), symbol)
        # print(f"fetch_symbol_price_ticker.binance_res: {binance_res}")
        return binance_res

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#symbol-order-book-ticker
    async def fetch_symbol_order_book_ticker(self, symbol=None):
        if symbol:
            self.assert_symbol_exists(symbol)
        return await self.http.send_api_call(
            "/api/v3/ticker/bookTicker",
            params={"symbol": symbol} if symbol else {},
            signed=False,
            send_api_key=False,
        )

    # ACCOUNT ENDPOINTS

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#new-order--trade
    async def create_order(
        self,
        symbol,
        side,
        order_type,
        time_in_force=None,
        quantity=None,
        quote_order_quantity=None,
        price=None,
        new_client_order_id=None,
        stop_price=None,
        iceberg_quantity=None,
        response_type=None,
        receive_window=None,
        test=False,
    ):
        self.assert_symbol(symbol)
        side = self.enum_to_value(side)
        order_type = self.enum_to_value(order_type)
        if not side:
            raise ValueError("This query requires a side.")
        if not order_type:
            raise ValueError("This query requires an order_type.")

        binance_res = {}
        if self.exchange == 'binance':
            params = {"symbol": symbol, "side": side, "type": order_type}
            if time_in_force:
                params["timeInForce"] = self.enum_to_value(time_in_force)
            elif order_type in [
                OrderType.LIMIT.value,
                OrderType.STOP_LOSS_LIMIT.value,
                OrderType.TAKE_PROFIT_LIMIT.value,
            ]:
                raise ValueError("This order type requires a time_in_force.")
            if quote_order_quantity:
                params["quoteOrderQty"] = self.refine_amount(
                    symbol, quote_order_quantity, True
                )
            if quantity:
                params["quantity"] = self.refine_amount(symbol, quantity)
            elif not quote_order_quantity:
                raise ValueError(
                    "This order type requires a quantity or a quote_order_quantity."
                    if order_type == OrderType.MARKET
                    else "This order type requires a quantity."
                )
            if price:
                params["price"] = self.refine_price(symbol, price)
            elif order_type in [
                OrderType.LIMIT.value,
                OrderType.STOP_LOSS_LIMIT.value,
                OrderType.TAKE_PROFIT_LIMIT.value,
                OrderType.LIMIT_MAKER.value,
            ]:
                raise ValueError("This order type requires a price.")
            if new_client_order_id:
                params["newClientOrderId"] = new_client_order_id
            if stop_price:
                params["stopPrice"] = self.refine_price(symbol, stop_price)
            elif order_type in [
                OrderType.STOP_LOSS.value,
                OrderType.STOP_LOSS_LIMIT.value,
                OrderType.TAKE_PROFIT.value,
                OrderType.TAKE_PROFIT_LIMIT.value,
            ]:
                raise ValueError("This order type requires a stop_price.")
            if iceberg_quantity:
                params["icebergQty"] = self.refine_amount(symbol, iceberg_quantity)
            if response_type:
                params["newOrderRespType"] = response_type
            if receive_window:
                params["recvWindow"] = receive_window
            route = "/api/v3/order/test" if test else "/api/v3/order"
            binance_res = await self.http.send_api_call(route, "POST", data=params, signed=True)
        elif self.exchange == 'ftx':
            params = {
                "market": self.symbol_to_ftx(symbol),
                "side": side.lower(),
                "price": float(price),
                "type": order_type.lower(),
                "size": float(quantity),
                "clientId": None
            }
            count = 0
            res = {}
            while True and count < 10:
                try:
                    res = await self.http.send_api_call(
                        "orders",
                        method="POST",
                        signed=True,
                        **params,
                    )
                    break
                except RateLimitReached:
                    logger.info(f"RateLimitReached for {self.symbol_to_ftx(symbol)}, count {count}, try one else")
                    count += 1
                    await asyncio.sleep(random.uniform(0.1, 0.3) * count)
            # logger.debug(f"create_order.res: {res}")
            if res and res.get('success'):
                binance_res = ftx_order(res.get('result'), response_type=False)
                if binance_res.get('status') != 'NEW':
                    order_id = binance_res.get('orderId')
                    binance_res = await self.fetch_order(symbol, order_id, receive_window)
        return binance_res

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#query-order-user_data
    async def fetch_order(  # lgtm [py/similar-function]
        self,
        symbol,
        order_id=None,
        origin_client_order_id=None,
        receive_window=None,
        response_type=None,
    ):
        self.assert_symbol(symbol)
        binance_res = {}
        if self.exchange == 'binance':
            params = {"symbol": symbol}
            if not order_id and not origin_client_order_id:
                raise ValueError(
                    "This query requires an order_id or an origin_client_order_id"
                )
            if order_id:
                params["orderId"] = order_id
            if origin_client_order_id:
                params["originClientOrderId"] = origin_client_order_id
            if receive_window:
                params["recvWindow"] = receive_window

            binance_res = await self.http.send_api_call(
                "/api/v3/order",
                params=params,
                signed=True,
            )
        elif self.exchange == 'ftx':
            if not order_id:
                raise ValueError(
                    "This query requires an order_id on FTX"
                )
            res = await self.http.send_api_call(
                f"orders/{order_id}",
                signed=True,
             )
            # print(f"fetch_order.res: {res}")
            if res and res.get('success'):
                binance_res = ftx_order(res.get('result'), response_type=response_type)
        # print(f"fetch_order.binance_res: {binance_res}")
        return binance_res

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#cancel-order-trade
    async def cancel_order(  # lgtm [py/similar-function]
        self,
        symbol,
        order_id=None,
        origin_client_order_id=None,
        new_client_order_id=None,
        receive_window=None,
    ):
        self.assert_symbol(symbol)
        binance_res = {}
        if self.exchange == 'binance':
            params = {"symbol": symbol}
            if not order_id and not origin_client_order_id:
                raise ValueError(
                    "This query requires an order_id or an origin_client_order_id."
                )
            if order_id:
                params["orderId"] = order_id
            if origin_client_order_id:
                params["originClientOrderId"] = origin_client_order_id
            if new_client_order_id:
                params["newClientOrderId"] = origin_client_order_id
            if receive_window:
                params["recvWindow"] = receive_window
            binance_res = await self.http.send_api_call(
                "/api/v3/order",
                "DELETE",
                params=params,
                signed=True,
            )
        elif self.exchange == 'ftx':
            if not order_id:
                raise ValueError(
                    "This query requires an order_id on FTX"
                )
            await self.http.send_api_call(
                f"orders/{order_id}",
                method="DELETE",
                signed=True,
             )
            order_not_cancelled = True
            timeout = STATUS_TIMEOUT
            while order_not_cancelled and timeout:
                timeout -= 1
                binance_res = await self.fetch_order(symbol,
                                                     order_id,
                                                     origin_client_order_id,
                                                     receive_window,
                                                     response_type=True,
                                                     )
                order_not_cancelled = False if binance_res.get('status') == 'CANCELED' else True
                await asyncio.sleep(1)
        return binance_res

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#cancel-all-open-orders-on-a-symbol-trade
    async def cancel_all_orders(self, symbol, receive_window=None):
        self.assert_symbol(symbol)
        binance_res = []
        if self.exchange == 'binance':
            params = {"symbol": symbol}
            if receive_window:
                params["recvWindow"] = receive_window
            binance_res = await self.http.send_api_call(
                "/api/v3/openOrders",
                "DELETE",
                params=params,
                signed=True,
            )
        elif self.exchange == 'ftx':
            params = {'market': self.symbol_to_ftx(symbol)}
            # Get list of open orders
            res_orders = await self.http.send_api_call(
                "orders",
                signed=True,
                **params
            )
            # print(f"cancel_all_orders.res_orders: {res_orders}")
            # Delete it
            res = await self.http.send_api_call(
                "orders",
                method="DELETE",
                signed=True,
                **params
            )
            # print(f"cancel_all_orders.res: {res}")
            if res and res.get('success'):
                binance_res = ftx_orders(res_orders.get('result'), response_type=True)
            # print(f"cancel_all_orders.binance_res: {binance_res}")
        return binance_res

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#current-open-orders-user_data
    async def fetch_open_orders(self, symbol, receive_window=None):
        self.assert_symbol(symbol)
        binance_res = []
        if self.exchange == 'binance':
            params = {"symbol": symbol}
            if receive_window:
                params["recvWindow"] = receive_window
            binance_res = await self.http.send_api_call(
                "/api/v3/openOrders",
                params=params,
                signed=True
            )
        elif self.exchange == 'ftx':
            params = {'market': self.symbol_to_ftx(symbol)}
            res = await self.http.send_api_call(
                "orders",
                signed=True,
                **params,
            )
            # print(f"fetch_open_orders.res: {res}")
            if res and res.get('success'):
                binance_res = ftx_orders(res.get('result'))
        # print(f"fetch_open_orders.binance_res: {binance_res}")
        return binance_res

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#all-orders-user_data
    async def fetch_all_orders(
        self,
        symbol,
        order_id=None,
        start_time=None,
        end_time=None,
        limit=500,
        receive_window=None,
    ):
        self.assert_symbol(symbol)
        if limit == 500:
            params = {"symbol": symbol}
        elif 0 < limit <= 1000:
            params = {"symbol": symbol, "limit": limit}
        else:
            raise ValueError(
                f"{limit} is not a valid limit. A valid limit should be > 0 and <= to 1000."
            )

        if order_id:
            params["orderId"] = order_id
        if start_time:
            params["startTime"] = start_time
        if end_time:
            params["endTime"] = end_time
        if receive_window:
            params["recvWindow"] = receive_window

        return await self.http.send_api_call(
            "/api/v3/allOrders",
            params=params,
            signed=True,
        )

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#new-oco-trade
    async def create_oco(
        self,
        symbol,
        side,
        quantity,
        price,
        stop_price,
        list_client_order_id=None,
        limit_iceberg_quantity=None,
        stop_client_order_id=None,
        stop_limit_price=None,
        stop_iceberg_quantity=None,
        stop_limit_time_in_force=None,
        response_type=None,
        receive_window=None,
    ):
        self.assert_symbol(symbol)
        side = self.enum_to_value(side)
        if not side:
            raise ValueError("This query requires a side.")
        if not quantity:
            raise ValueError("This query requires a quantity.")
        if not price:
            raise ValueError("This query requires a price.")
        if not stop_price:
            raise ValueError("This query requires a stop_price.")

        params = {
            "symbol": symbol,
            "side": side,
            "quantity": self.refine_amount(symbol, quantity),
            "price": self.refine_price(symbol, price),
            "stopPrice": self.refine_price(symbol, stop_price),
            "stopLimitPrice": self.refine_price(symbol, stop_limit_price),
        }

        if list_client_order_id:
            params["listClientOrderId"] = list_client_order_id
        if limit_iceberg_quantity:
            params["limitIcebergQty"] = self.refine_amount(
                symbol, limit_iceberg_quantity
            )
        if stop_client_order_id:
            params["stopLimitPrice"] = self.refine_price(symbol, stop_client_order_id)
        if stop_iceberg_quantity:
            params["stopIcebergQty"] = self.refine_amount(symbol, stop_iceberg_quantity)
        if stop_limit_time_in_force:
            params["stopLimitTimeInForce"] = stop_limit_time_in_force
        if response_type:
            params["newOrderRespType"] = response_type
        if receive_window:
            params["recvWindow"] = receive_window

        return await self.http.send_api_call(
            "/api/v3/order/oco", "POST", data=params, signed=True
        )

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#query-oco-user_data
    async def fetch_oco(  # lgtm [py/similar-function]
        self,
        symbol,
        order_list_id=None,
        origin_client_order_id=None,
        receive_window=None,
    ):
        self.assert_symbol(symbol)
        params = {"symbol": symbol}
        if not order_list_id and not origin_client_order_id:
            raise ValueError(
                "This query requires an order_id or an origin_client_order_id."
            )
        if order_list_id:
            params["orderListId"] = order_list_id
        if origin_client_order_id:
            params["originClientOrderId"] = origin_client_order_id
        if receive_window:
            params["recvWindow"] = receive_window

        return await self.http.send_api_call(
            "/api/v3/orderList",
            params=params,
            signed=True,
        )

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#cancel-oco-trade
    async def cancel_oco(  # lgtm [py/similar-function]
        self,
        symbol,
        order_list_id=None,
        list_client_order_id=None,
        new_client_order_id=None,
        receive_window=None,
    ):
        self.assert_symbol(symbol)
        params = {"symbol": symbol}
        if not order_list_id and not list_client_order_id:
            raise ValueError(
                "This query requires a order_list_id or a list_client_order_id."
            )
        if order_list_id:
            params["orderListId"] = order_list_id
        if list_client_order_id:
            params["listClientOrderId"] = list_client_order_id
        if new_client_order_id:
            params["newClientOrderId"] = new_client_order_id
        if receive_window:
            params["recvWindow"] = receive_window

        return await self.http.send_api_call(
            "/api/v3/order/oco",
            "DELETE",
            params=params,
            signed=True,
        )

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#query-open-oco-user_data
    async def fetch_open_oco(self, receive_window=None):
        params = {}

        if receive_window:
            params["recvWindow"] = receive_window

        return await self.http.send_api_call(
            "/api/v3/openOrderList",
            params=params,
            signed=True,
        )

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#query-all-oco-user_data
    async def fetch_all_oco(
        self,
        from_id=None,
        start_time=None,
        end_time=None,
        limit=None,
        receive_window=None,
    ):
        params = {}

        if from_id:
            params["fromId"] = from_id
        if start_time:
            params["startTime"] = start_time
        if end_time:
            params["endTime"] = end_time
        if limit:
            params["limit"] = limit
        if receive_window:
            params["recvWindow"] = receive_window

        return await self.http.send_api_call(
            "/api/v3/allOrderList",
            params=params,
            signed=True,
        )

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#account-information-user_data
    async def fetch_account_information(self, receive_window=None):
        params = {}
        binance_res = {}
        if self.exchange == 'binance':
            if receive_window:
                params["recvWindow"] = receive_window
            binance_res = await self.http.send_api_call(
                "/api/v3/account",
                params=params,
                signed=True,
            )
        elif self.exchange == 'ftx':
            res = await self.http.send_api_call(
                "wallet/balances",
                signed=True)
            # print(f"fetch_account_information.res: {res}")
            if res and res.get('success'):
                binance_res = ftx_account_information(res.get('result'))
        # print(f"fetch_account_information.binance_res: {binance_res}")
        return binance_res

    # https://binance-docs.github.io/apidocs/spot/en/#funding-wallet-user_data
    # Not can be used for Spot Test Network, for real SPOT market only
    async def fetch_funding_wallet(self, asset=None, need_btc_valuation=None, receive_window=None):
        binance_res = []
        if self.exchange == 'binance':
            params = {}
            if asset:
                params["asset"] = asset
            if need_btc_valuation:
                params["needBtcValuation"] = "true"
            if receive_window:
                params["recvWindow"] = receive_window
            binance_res = await self.http.send_api_call(
                "/sapi/v1/asset/get-funding-asset",
                method="POST",
                params=params,
                signed=True,
            )
        elif self.exchange == 'ftx':
            res = await self.http.send_api_call(
                "wallet/all_balances",
                signed=True,
             )
            # print(f"fetch_funding_wallet.res: {res}")
            if res and res.get('success'):
                binance_res = ftx_fetch_funding_wallet(res.get('result').get('main', []))
        # print(f"fetch_funding_wallet.binance_res: {binance_res}")
        return binance_res

    # https://github.com/binance/binance-spot-api-docs/blob/master/rest-api.md#account-trade-list-user_data
    async def fetch_account_trade_list(
        self,
        symbol,
        start_time=None,
        end_time=None,
        from_id=None,
        limit=500,
        receive_window=None,
    ):
        self.assert_symbol(symbol)
        binance_res = []
        if self.exchange == 'binance':
            if limit == 500:
                params = {"symbol": symbol}
            elif 0 < limit <= 1000:
                params = {"symbol": symbol, "limit": limit}
            else:
                raise ValueError(
                    f"{limit} is not a valid limit. A valid limit should be > 0 and <= to 1000."
                )
            if start_time:
                params["startTime"] = start_time
            if end_time:
                params["endTime"] = end_time
            if from_id:
                params["fromId"] = from_id
            if receive_window:
                params["recvWindow"] = receive_window
            binance_res = await self.http.send_api_call(
                "/api/v3/myTrades",
                params=params,
                signed=True,
            )
        elif self.exchange == 'ftx':
            params = {'market': self.symbol_to_ftx(symbol),
                      'start_time': int(start_time / 1000)}
            if from_id:
                params["orderId"] = from_id
            res = await self.http.send_api_call(
                "fills",
                signed=True,
                **params,
            )
            # print(f"fetch_account_trade_list.res: {res}")
            if res and res.get('success'):
                binance_res = ftx_account_trade_list(res.get('result')[-limit:])
        # print(f"fetch_account_trade_list.binance_res: {binance_res}")
        return binance_res
    # USER DATA STREAM ENDPOINTS

    # https://github.com/binance-exchange/binance-official-api-docs/blob/master/user-data-stream.md#create-a-listenkey
    async def create_listen_key(self):
        return await self.http.send_api_call("/api/v3/userDataStream", "POST")

    # https://github.com/binance-exchange/binance-official-api-docs/blob/master/user-data-stream.md#close-a-listenkey
    async def keep_alive_listen_key(self, listen_key):
        if not listen_key:
            raise ValueError("This query requires a listen_key.")
        return await self.http.send_api_call(
            "/api/v3/userDataStream", "PUT", params={"listenKey": listen_key}
        )

    # https://github.com/binance-exchange/binance-official-api-docs/blob/master/user-data-stream.md#close-a-listenkey
    async def close_listen_key(self, listen_key):
        if not listen_key:
            raise ValueError("This query requires a listen_key.")
        return await self.http.send_api_call(
            "/api/v3/userDataStream", "DELETE", params={"listenKey": listen_key}
        )



#!/usr/bin/python3.8
# -*- coding: utf-8 -*-
"""
Client example for exchanges-wrapper exch_srv.py
"""

import asyncio
import toml
# noinspection PyPackageRequirements
import grpc
# noinspection PyPackageRequirements
from google.protobuf import json_format
import api_pb2
import api_pb2_grpc

# Set id exchange from ms_cfg.toml
ID_EXCHANGE = 0

# For more channel options, please see https://grpc.io/grpc/core/group__grpc__arg__keys.html
CHANNEL_OPTIONS = [('grpc.lb_policy_name', 'pick_first'),
                   ('grpc.enable_retries', 0),
                   ('grpc.keepalive_timeout_ms', 10000)]
HEARTBEAT = 2  # Sec
RATE_LIMITER = HEARTBEAT * 5
FILE_CONFIG = 'ms_cfg.toml'
config = toml.load(FILE_CONFIG)
EXCHANGE = config.get('exchange')
SYMBOL = 'BTCUSDT'


async def main(_symbol):
    account_name = EXCHANGE[ID_EXCHANGE]
    print(f"main.account_name: {account_name}")  # lgtm [py/clear-text-logging-sensitive-data]
    channel = grpc.aio.insecure_channel(target='localhost:50051', options=CHANNEL_OPTIONS)
    stub = api_pb2_grpc.MartinStub(channel)
    client_id_msg = await stub.OpenClientConnection(api_pb2.OpenClientConnectionRequest(
        account_name=account_name,
        rate_limiter=RATE_LIMITER))
    print(f"main.client_id: {client_id_msg.client_id}")
    print(f"main.srv_version: {client_id_msg.srv_version}")
    #
    _exchange_info_symbol = await stub.FetchExchangeInfoSymbol(api_pb2.MarketRequest(
        client_id=client_id_msg.client_id,
        symbol=_symbol))
    exchange_info_symbol = json_format.MessageToDict(_exchange_info_symbol)
    print("\n".join(f"{k}\t{v}" for k, v in exchange_info_symbol.items()))


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main(SYMBOL))
    loop.stop()
    loop.close()



#!/usr/bin/python3.8
# -*- coding: utf-8 -*-

from exchanges_wrapper import __version__

import time
import weakref
# from pympler import tracker
import gc

import asyncio
import functools
import json
# import logging
import logging.handlers
import os
import sys
# from datetime import datetime

# noinspection PyPackageRequirements
import grpc
import toml
# noinspection PyPackageRequirements
from google.protobuf import json_format

import events
import errors
import ftx_parser as ftx
from client import Client
from definitions import Side, OrderType, TimeInForce, ResponseType
import api_pb2
import api_pb2_grpc
#
#
sys.tracebacklimit = 0
FILE_CONFIG = 'exch_srv_cfg.toml'
CONFIG = None
if os.path.exists(FILE_CONFIG):
    CONFIG = toml.load(FILE_CONFIG)
else:
    print("Can't find config file!")
    # noinspection PyProtectedMember,PyUnresolvedReferences
    os._exit(1)
HEARTBEAT = 1  # Sec


def get_account(_account_name: str) -> ():
    # print(f"get_account")
    accounts = CONFIG.get('accounts')
    sub_account = None
    for account in accounts:
        # print(f"get_account.account: {account}")
        if account.get('name') == _account_name:
            if account.get('test_net'):
                real_market = False
                api = CONFIG.get('endpoint_test').get('api')
                ws = CONFIG.get('endpoint_test').get('ws')
            else:
                if 'FTX' in _account_name:
                    real_market = True
                    api = CONFIG.get('endpoint_ftx').get('api')
                    ws = CONFIG.get('endpoint_ftx').get('ws')
                    sub_account = account.get('sub_account_name')
                else:
                    real_market = True
                    api = CONFIG.get('endpoint').get('api')
                    ws = CONFIG.get('endpoint').get('ws')
            return account.get('api_key'), account.get('api_secret'), api, ws, real_market, sub_account
    return ()


class OpenClient:
    open_clients = []

    def __init__(self, _account_name: str):
        # lgtm [py/mismatched-multiple-assignment]
        api_key, api_secret, api, ws, real_mrkt, sub_account = get_account(_account_name)
        self.name = _account_name
        self.real_market = real_mrkt
        self.client = Client(api_key, api_secret, endpoint=api, sub_account=sub_account, endpoint_ws=ws)
        self.stop_streams_for_symbol = None
        self.stream_queue = []
        self.on_order_update_queue = asyncio.Queue()
        OpenClient.open_clients.append(self)

    @classmethod
    def get_id(cls, _account_name):
        _id = 0
        for open_client in cls.open_clients:
            if open_client.name == _account_name:
                _id = id(open_client)
                break
        return _id

    @classmethod
    def get_client(cls, _id):
        _client = None
        for open_client in cls.open_clients:
            if id(open_client) == _id:
                _client = open_client
                break
        return _client


class Event:
    def __init__(self, event_data: {}):
        self.symbol = event_data["symbol"]
        self.client_order_id = event_data["clientOrderId"]
        self.side = event_data["side"]
        self.order_type = event_data["type"]
        self.time_in_force = event_data["timeInForce"]
        self.order_quantity = event_data["origQty"]
        self.order_price = event_data["price"]
        self.stop_price = event_data["stopPrice"]
        self.iceberg_quantity = event_data["icebergQty"]
        self.order_list_id = event_data["orderListId"]
        self.original_client_id = event_data["clientOrderId"]
        self.execution_type = "TRADE"
        self.order_status = event_data["status"]
        self.order_reject_reason = "NONE"
        self.order_id = event_data["orderId"]
        self.last_executed_quantity = event_data["executedQty"]
        self.cumulative_filled_quantity = event_data["executedQty"]
        self.last_executed_price = event_data["price"]
        self.commission_amount = "0.0"
        self.commission_asset = ""
        self.transaction_time = event_data["updateTime"]
        self.trade_id = 1
        self.ignore_a = int()
        self.in_order_book = False
        self.is_maker_side = False
        self.ignore_b = False
        self.order_creation_time = event_data["time"]
        self.quote_asset_transacted = event_data["cummulativeQuoteQty"]
        self.last_quote_asset_transacted = event_data["cummulativeQuoteQty"]
        self.quote_order_quantity = event_data["origQuoteOrderQty"]


# noinspection PyPep8Naming,PyMethodMayBeStatic,PyUnusedLocal
class Martin(api_pb2_grpc.MartinServicer):
    rate_limit_reached_time = None
    rate_limiter = None

    async def OpenClientConnection(self, request: api_pb2.OpenClientConnectionRequest,
                                   context: grpc.aio.ServicerContext) -> api_pb2.OpenClientConnectionId:
        # logger.info(f"OpenClientConnection: {request.account_name}")
        client_id = OpenClient.get_id(request.account_name)
        # print(f"OpenClientConnection.request.account_name: {request.account_name}")
        if not client_id and get_account(request.account_name):
            open_client = OpenClient(request.account_name)
            # print(f"OpenClientConnection.open_client.name: {open_client.name}")
            if open_client:
                await open_client.client.load()
                client_id = id(open_client)
        # Set rate_limiter
        Martin.rate_limiter = max(Martin.rate_limiter if Martin.rate_limiter else 0, request.rate_limiter)
        return api_pb2.OpenClientConnectionId(client_id=client_id, srv_version=__version__)

    async def FetchServerTime(self, request: api_pb2.OpenClientConnectionId,
                              context: grpc.aio.ServicerContext) -> api_pb2.FetchServerTimeResponse:
        client = OpenClient.get_client(request.client_id).client
        res = await client.fetch_server_time()
        server_time = res.get('serverTime')
        return api_pb2.FetchServerTimeResponse(server_time=server_time)

    async def ResetRateLimit(self, request: api_pb2.OpenClientConnectionId,
                             context: grpc.aio.ServicerContext) -> api_pb2.SimpleResponse:
        Martin.rate_limiter = max(Martin.rate_limiter if Martin.rate_limiter else 0, request.rate_limiter)
        _success = False
        client = OpenClient.get_client(request.client_id).client
        if Martin.rate_limit_reached_time:
            if time.time() - Martin.rate_limit_reached_time > 30:
                client.http.rate_limit_reached = False
                Martin.rate_limit_reached_time = None
                logger.info("ResetRateLimit error clear, trying one else time")
                _success = True
        else:
            if client.http.rate_limit_reached:
                Martin.rate_limit_reached_time = time.time()
        return api_pb2.SimpleResponse(success=_success)

    async def FetchOpenOrders(self, request: api_pb2.MarketRequest,
                              context: grpc.aio.ServicerContext) -> api_pb2.FetchOpenOrdersResponse:
        open_client = OpenClient.get_client(request.client_id)
        client = open_client.client
        # message list
        response = api_pb2.FetchOpenOrdersResponse()
        # Nested dict
        response_order = api_pb2.FetchOpenOrdersResponse.Order()
        try:
            res = await client.fetch_open_orders(symbol=request.symbol, receive_window=None)
        except asyncio.CancelledError:
            pass  # Task cancellation should not be logged as an error
        except errors.RateLimitReached as ex:
            Martin.rate_limit_reached_time = time.time()
            logger.warning(f"FetchOpenOrders for {open_client.name}:{request.symbol} exception: {ex}")
            context.set_details(f"{ex}")
            context.set_code(grpc.StatusCode.RESOURCE_EXHAUSTED)
        except errors.HTTPError as ex:
            logger.error(f"FetchOpenOrders for {open_client.name}:{request.symbol} exception: {ex}")
            context.set_details(f"{ex}")
            context.set_code(grpc.StatusCode.FAILED_PRECONDITION)
        except Exception as ex:
            logger.error(f"FetchOpenOrders for {open_client.name}:{request.symbol} exception: {ex}")
            context.set_details(f"{ex}")
            context.set_code(grpc.StatusCode.UNKNOWN)
        else:
            # logger.info(f"FetchOpenOrders.res: {res}")
            for order in res:
                new_order = json_format.ParseDict(order, response_order)
                # logger.info(f"FetchOpenOrders.new_order: {new_order}")
                response.items.append(new_order)
        response.rate_limiter = Martin.rate_limiter
        return response

    async def FetchOrder(self, request: api_pb2.FetchOrderRequest,
                         context: grpc.aio.ServicerContext) -> api_pb2.FetchOrderResponse:
        open_client = OpenClient.get_client(request.client_id)
        client = open_client.client
        _queue = open_client.on_order_update_queue
        response = api_pb2.FetchOrderResponse()
        try:
            res = await client.fetch_order(symbol=request.symbol,
                                           order_id=request.order_id,
                                           origin_client_order_id=None,
                                           receive_window=None)
        except asyncio.CancelledError:
            pass  # Task cancellation should not be logged as an error
        except Exception as _ex:
            logger.info(f"FetchOrders for {open_client.name}:{request.symbol} exception: {_ex}")
        else:
            # logger.info(f"FetchOrder: {res}")
            if request.filled_update_call and res.get('status') == 'FILLED':
                logger.info(f"FetchOrder: {res.get('status')}")
                event = Event(res)
                logger.info(f"FetchOrder.event: {open_client.name}:{event.symbol}:{int(event.order_id)}:"
                            f"{event.order_status}")
                _event = weakref.ref(event)
                await _queue.put(_event())
            json_format.ParseDict(res, response)
        return response

    async def CancelAllOrders(self, request: api_pb2.MarketRequest,
                              context: grpc.aio.ServicerContext) -> api_pb2.CancelAllOrdersResponse:
        client = OpenClient.get_client(request.client_id).client
        # message list
        response = api_pb2.CancelAllOrdersResponse()
        # Nested dict
        response_order = api_pb2.CancelAllOrdersResponse.CancelOrder()
        res = await client.cancel_all_orders(symbol=request.symbol, receive_window=None)
        # logger.info(f"CancelAllOrders: {res}")
        for order in res:
            cancel_order = json_format.ParseDict(order, response_order)
            response.items.append(cancel_order)
        return response

    async def FetchExchangeInfoSymbol(self, request: api_pb2.MarketRequest,
                                      context: grpc.aio.ServicerContext
                                      ) -> api_pb2.FetchExchangeInfoSymbolResponse:
        client = OpenClient.get_client(request.client_id).client
        response = api_pb2.FetchExchangeInfoSymbolResponse()
        exchange_info = await client.fetch_exchange_info()
        exchange_info_symbol = {}
        try:
            exchange_info_symbol = next(item for item in exchange_info.get('symbols')
                                        if item["symbol"] == request.symbol)
        except StopIteration:
            logger.info("FetchExchangeInfoSymbol.exchange_info_symbol: None")
        # logger.info(f"exchange_info_symbol: {exchange_info_symbol}")
        filters_res = exchange_info_symbol.pop('filters', [])
        json_format.ParseDict(exchange_info_symbol, response)
        # logger.info(f"filters: {filters_res}")
        filters = response.filters
        for _filter in filters_res:
            if _filter.get('filterType') == 'PRICE_FILTER':
                new_filter_template = api_pb2.FetchExchangeInfoSymbolResponse.Filters.PriceFilter()
                new_filter = json_format.ParseDict(_filter, new_filter_template)
                filters.price_filter.CopyFrom(new_filter)
            elif _filter.get('filterType') == 'PERCENT_PRICE':
                new_filter_template = api_pb2.FetchExchangeInfoSymbolResponse.Filters.PercentPrice()
                new_filter = json_format.ParseDict(_filter, new_filter_template)
                filters.percent_price.CopyFrom(new_filter)
            elif _filter.get('filterType') == 'LOT_SIZE':
                new_filter_template = api_pb2.FetchExchangeInfoSymbolResponse.Filters.LotSize()
                new_filter = json_format.ParseDict(_filter, new_filter_template)
                filters.lot_size.CopyFrom(new_filter)
            elif _filter.get('filterType') == 'MIN_NOTIONAL':
                new_filter_template = api_pb2.FetchExchangeInfoSymbolResponse.Filters.MinNotional()
                new_filter = json_format.ParseDict(_filter, new_filter_template)
                filters.min_notional.CopyFrom(new_filter)
            elif _filter.get('filterType') == 'ICEBERG_PARTS':
                new_filter_template = api_pb2.FetchExchangeInfoSymbolResponse.Filters.IcebergParts()
                new_filter = json_format.ParseDict(_filter, new_filter_template)
                filters.iceberg_parts.CopyFrom(new_filter)
            elif _filter.get('filterType') == 'MARKET_LOT_SIZE':
                new_filter_template = api_pb2.FetchExchangeInfoSymbolResponse.Filters.MarketLotSize()
                new_filter = json_format.ParseDict(_filter, new_filter_template)
                filters.market_lot_size.CopyFrom(new_filter)
            elif _filter.get('filterType') == 'MAX_NUM_ORDERS':
                new_filter_template = api_pb2.FetchExchangeInfoSymbolResponse.Filters.MaxNumOrders()
                new_filter = json_format.ParseDict(_filter, new_filter_template)
                filters.max_num_orders.CopyFrom(new_filter)
            elif _filter.get('filterType') == 'MAX_NUM_ICEBERG_ORDERS':
                new_filter_template = api_pb2.FetchExchangeInfoSymbolResponse.Filters.MaxNumIcebergOrders()
                new_filter = json_format.ParseDict(_filter, new_filter_template)
                filters.max_num_iceberg_orders.CopyFrom(new_filter)
            elif _filter.get('filterType') == 'MAX_POSITION':
                new_filter_template = api_pb2.FetchExchangeInfoSymbolResponse.Filters.MaxPosition()
                new_filter = json_format.ParseDict(_filter, new_filter_template)
                filters.max_position.CopyFrom(new_filter)
        return response

    async def FetchAccountInformation(self, request: api_pb2.OpenClientConnectionId,
                                      context: grpc.aio.ServicerContext
                                      ) -> api_pb2.FetchAccountBalanceResponse:
        open_client = OpenClient.get_client(request.client_id)
        client = open_client.client
        response = api_pb2.FetchAccountBalanceResponse()
        response_balance = api_pb2.FetchAccountBalanceResponse.Balances()
        account_information = await client.fetch_account_information(receive_window=None)
        # logger.info(f"account_information: {account_information}")
        # Send only balances
        res = account_information.get('balances', [])
        # Create consolidated list of asset balances from SPOT and Funding wallets
        balances = []
        for i in res:
            _free = float(i.get('free'))
            _locked = float(i.get('locked'))
            if _free or _locked:
                balances.append({'asset': i.get('asset'), 'free': i.get('free'), 'locked': i.get('locked')})
        # logger.info(f"account_information.balances: {balances}")
        for balance in balances:
            new_balance = json_format.ParseDict(balance, response_balance)
            response.balances.extend([new_balance])
        return response

    async def FetchFundingWallet(self, request: api_pb2.FetchFundingWalletRequest,
                                 context: grpc.aio.ServicerContext) -> api_pb2.FetchFundingWalletResponse:
        open_client = OpenClient.get_client(request.client_id)
        client = open_client.client
        response = api_pb2.FetchFundingWalletResponse()
        response_balance = api_pb2.FetchFundingWalletResponse.Balances()
        res = []
        if open_client.real_market and client.exchange in ('binance', 'ftx'):
            try:
                res = await client.fetch_funding_wallet(asset=request.asset,
                                                        need_btc_valuation=request.need_btc_valuation,
                                                        receive_window=request.receive_window)
            except AttributeError:
                logger.error("Can't get Funding Wallet balances")
        # logger.info(f"funding_wallet: {res}")
        for balance in res:
            new_balance = json_format.ParseDict(balance, response_balance)
            response.balances.extend([new_balance])
        return response

    async def FetchOrderBook(self, request: api_pb2.MarketRequest,
                             context: grpc.aio.ServicerContext) -> api_pb2.FetchOrderBookResponse:
        client = OpenClient.get_client(request.client_id).client
        response = api_pb2.FetchOrderBookResponse()
        res = await client.fetch_order_book(symbol=request.symbol, limit=5)
        res_bids = res.get('bids', [])
        res_asks = res.get('asks', [])
        response.lastUpdateId = res.get('lastUpdateId')
        for bid in res_bids:
            response.bids.append(json.dumps(bid))
        for ask in res_asks:
            response.asks.append(json.dumps(ask))
        return response

    async def FetchSymbolPriceTicker(
            self, request: api_pb2.MarketRequest,
            context: grpc.aio.ServicerContext) -> api_pb2.FetchSymbolPriceTickerResponse:
        client = OpenClient.get_client(request.client_id).client
        response = api_pb2.FetchSymbolPriceTickerResponse()
        res = await client.fetch_symbol_price_ticker(symbol=request.symbol)
        json_format.ParseDict(res, response)
        return response

    async def FetchTickerPriceChangeStatistics(
            self, request: api_pb2.MarketRequest,
            context: grpc.aio.ServicerContext) -> api_pb2.FetchTickerPriceChangeStatisticsResponse:
        client = OpenClient.get_client(request.client_id).client
        response = api_pb2.FetchTickerPriceChangeStatisticsResponse()
        res = await client.fetch_ticker_price_change_statistics(symbol=request.symbol)
        json_format.ParseDict(res, response)
        return response

    async def FetchKlines(self, request: api_pb2.FetchKlinesRequest,
                          context: grpc.aio.ServicerContext) -> api_pb2.FetchKlinesResponse:
        client = OpenClient.get_client(request.client_id).client
        response = api_pb2.FetchKlinesResponse()
        try:
            res = await client.fetch_klines(symbol=request.symbol, interval=request.interval,
                                            start_time=None, end_time=None, limit=request.limit)
        except asyncio.CancelledError:
            pass  # Task cancellation should not be logged as an error
        except Exception as _ex:
            logger.info(f"FetchKlines for {request.symbol} interval: {request.interval}, exception: {_ex}")
        else:
            # logger.info(res)
            for candle in res:
                response.klines.append(json.dumps(candle))
        return response

    async def OnKlinesUpdate(self, request: api_pb2.FetchKlinesRequest,
                             context: grpc.aio.ServicerContext) -> api_pb2.OnKlinesUpdateResponse:
        response = api_pb2.OnKlinesUpdateResponse()
        open_client = OpenClient.get_client(request.client_id)
        client = open_client.client
        _queue = asyncio.Queue()
        open_client.stream_queue.append(_queue)
        open_client.stop_streams_for_symbol = str()
        _intervals = json.loads(request.interval)
        event_types = []
        # Register streams for intervals
        for i in _intervals:
            _event_type = f"{request.symbol.lower()}@kline_{i}"
            event_types.append(_event_type)
            client.events.register_event(functools.partial(on_klines_update, _queue), _event_type)
        while True:
            if open_client.stop_streams_for_symbol == request.symbol:
                [open_client.client.events.unregister(_event_type) for _event_type in event_types]
                logger.info(f"OnKlinesUpdate: Stop market stream for {open_client.name}:{request.symbol}:"
                            f"{_intervals}")
                break
            _event = await _queue.get()
            if _event:
                # logger.info(f"OnKlinesUpdate.event: {_event.symbol}:{_event.kline_interval}")
                response.symbol = _event.symbol
                response.interval = _event.kline_interval
                candle = [_event.kline_start_time,
                          _event.kline_open_price,
                          _event.kline_high_price,
                          _event.kline_low_price,
                          _event.kline_close_price,
                          _event.kline_base_asset_volume,
                          _event.kline_close_time,
                          _event.kline_quote_asset_volume,
                          _event.kline_trades_number,
                          _event.kline_taker_buy_base_asset_volume,
                          _event.kline_taker_buy_quote_asset_volume,
                          _event.kline_ignore
                          ]
                response.candle = json.dumps(candle)
                yield response

    async def FetchAccountTradeList(self, request: api_pb2.AccountTradeListRequest,
                                    context: grpc.aio.ServicerContext) -> api_pb2.AccountTradeListResponse:
        client = OpenClient.get_client(request.client_id).client
        response = api_pb2.AccountTradeListResponse()
        response_trade = api_pb2.AccountTradeListResponse.Trade()
        res = await client.fetch_account_trade_list(
            symbol=request.symbol,
            start_time=request.start_time,
            end_time=None,
            from_id=None,
            limit=request.limit,
            receive_window=None)
        # logger.info(f"FetchAccountTradeList: {res}")
        for trade in res:
            trade_order = json_format.ParseDict(trade, response_trade)
            response.items.append(trade_order)
        return response

    async def OnTickerUpdate(self, request: api_pb2.MarketRequest,
                             context: grpc.aio.ServicerContext) -> api_pb2.OnTickerUpdateResponse:
        response = api_pb2.OnTickerUpdateResponse()
        open_client = OpenClient.get_client(request.client_id)
        client = open_client.client
        _queue = asyncio.Queue()
        open_client.stream_queue.append(_queue)
        open_client.stop_streams_for_symbol = str()
        _symbol = client.symbol_to_ftx(request.symbol) if client.exchange == 'ftx' else request.symbol.lower()
        _event_type = f"{_symbol}@miniTicker"
        client.events.register_event(functools.partial(on_ticker_update, _queue), _event_type, client.exchange)
        while True:
            if open_client.stop_streams_for_symbol == request.symbol:
                open_client.client.events.unregister(_event_type, client.exchange)
                # logger.info(f"OnTickerUpdate: Stop market stream for {open_client.name}: {request.symbol}")
                break
            _event = await _queue.get()
            if _event:
                # logger.debug(f"OnTickerUpdate.event: {_event.symbol}, _event.close_price: {_event.close_price}")
                ticker_24h = {'symbol': _event.symbol,
                              'open_price': _event.open_price,
                              'close_price': _event.close_price,
                              'event_time': _event.event_time}
                json_format.ParseDict(ticker_24h, response)
                yield response

    async def OnOrderBookUpdate(self, request: api_pb2.MarketRequest,
                                context: grpc.aio.ServicerContext) -> api_pb2.FetchOrderBookResponse:
        response = api_pb2.FetchOrderBookResponse()
        open_client = OpenClient.get_client(request.client_id)
        client = open_client.client
        _queue = asyncio.Queue()
        open_client.stream_queue.append(_queue)
        open_client.stop_streams_for_symbol = str()
        _symbol = client.symbol_to_ftx(request.symbol) if client.exchange == 'ftx' else request.symbol.lower()
        _event_type = f"{_symbol}@depth5"
        client.events.register_event(functools.partial(on_order_book_update, _queue), _event_type, client.exchange)
        while True:
            if open_client.stop_streams_for_symbol == request.symbol:
                open_client.client.events.unregister(_event_type, client.exchange)
                logger.info(f"OnOrderBookUpdate: Stop market stream for {open_client.name}: {request.symbol}")
                break
            _event = await _queue.get()
            if _event:
                # logger.info(f"OnOrderBookUpdate._event: {_event}")
                response.Clear()
                response.lastUpdateId = _event.last_update_id
                for bid in _event.bids:
                    response.bids.append(json.dumps(bid))
                for ask in _event.asks:
                    response.asks.append(json.dumps(ask))
                yield response

    async def OnFundsUpdate(self, request: api_pb2.OnFundsUpdateRequest,
                            context: grpc.aio.ServicerContext) -> api_pb2.OnFundsUpdateResponse:
        response = api_pb2.OnFundsUpdateResponse()
        open_client = OpenClient.get_client(request.client_id)
        client = open_client.client
        open_client.stop_streams_for_symbol = str()
        if client.exchange == 'binance':
            _queue = asyncio.Queue()
            open_client.stream_queue.append(_queue)
            client.events.register_user_event(functools.partial(on_funds_update, _queue), 'outboundAccountPosition')
        balances_prev = []
        assets = [request.base_asset, request.quote_asset]
        while True:
            _event = None
            if open_client.stop_streams_for_symbol == request.symbol:
                client.events.unregister_user_event('outboundAccountPosition')
                logger.info(f"OnFundsUpdate: Stop user stream for {open_client.name}: {request.symbol}")
                break
            if client.exchange == 'binance':
                # noinspection PyUnboundLocalVariable
                _event = await _queue.get()
            elif client.exchange == 'ftx':
                await asyncio.sleep(HEARTBEAT * 3)
                try:
                    account_information = await client.fetch_account_information(receive_window=None)
                except (asyncio.exceptions.TimeoutError, errors.HTTPError, Exception) as _ex:
                    logger.info(f"OnFundsUpdate: for {open_client.name}"
                                f" {request.base_asset}/{request.quote_asset}: {_ex}")
                else:
                    balances = account_information.get('balances', {})
                    assets_balances = list(filter(lambda item: item['asset'] in assets, balances))
                    if assets_balances and assets_balances != balances_prev:
                        # logger.info(f"OnFundsUpdate.assets_balances: {assets_balances}")
                        content = ftx.ftx_on_funds_update(assets_balances)
                        balances_prev = assets_balances.copy()
                        _event = client.events.wrap_event(content)
            if _event:
                # logger.info(f"OnFundsUpdate: {_event.balances.items()}")
                response.funds = json.dumps(_event.balances)
                yield response

    async def OnOrderUpdate(self, request: api_pb2.MarketRequest,
                            context: grpc.aio.ServicerContext) -> api_pb2.OnOrderUpdateResponse:
        response = api_pb2.OnOrderUpdateResponse()
        open_client = OpenClient.get_client(request.client_id)
        client = open_client.client
        _queue = open_client.on_order_update_queue
        open_client.stream_queue.append(_queue)
        open_client.stop_streams_for_symbol = str()
        client.events.register_user_event(functools.partial(on_order_update, _queue), 'executionReport')
        while True:
            if open_client.stop_streams_for_symbol == request.symbol:
                client.events.unregister_user_event('executionReport')
                logger.debug(f"OnOrderUpdate: Stop user stream for {open_client.name}: {request.symbol}")
                break
            _event = await _queue.get()
            if _event:
                # logger.debug(f"OnOrderUpdate._event: {int(_event.order_id)}, {_event.order_status}")
                response.symbol = _event.symbol
                response.client_order_id = _event.client_order_id
                response.side = _event.side
                response.order_type = _event.order_type
                response.time_in_force = _event.time_in_force
                response.order_quantity = _event.order_quantity
                response.order_price = _event.order_price
                response.stop_price = _event.stop_price
                response.iceberg_quantity = _event.iceberg_quantity
                response.order_list_id = int(_event.order_list_id)
                response.original_client_id = _event.original_client_id
                response.execution_type = _event.execution_type
                response.order_status = _event.order_status
                response.order_reject_reason = _event.order_reject_reason
                response.order_id = int(_event.order_id)
                response.last_executed_quantity = _event.last_executed_quantity
                response.cumulative_filled_quantity = _event.cumulative_filled_quantity
                response.last_executed_price = _event.last_executed_price
                response.commission_amount = _event.commission_amount
                response.commission_asset = _event.commission_asset if _event.commission_asset else ''
                response.transaction_time = int(_event.transaction_time)
                response.trade_id = int(_event.trade_id)
                response.ignore_a = _event.ignore_a
                response.in_order_book = _event.in_order_book
                response.is_maker_side = bool(_event.is_maker_side)
                response.ignore_b = _event.ignore_b
                response.order_creation_time = int(_event.order_creation_time)
                response.quote_asset_transacted = _event.quote_asset_transacted
                response.last_quote_asset_transacted = _event.last_quote_asset_transacted
                response.quote_order_quantity = _event.quote_order_quantity
                yield response

    async def CreateLimitOrder(self, request: api_pb2.CreateLimitOrderRequest,
                               context: grpc.aio.ServicerContext) -> api_pb2.CreateLimitOrderResponse:
        response = api_pb2.CreateLimitOrderResponse()
        open_client = OpenClient.get_client(request.client_id)
        client = open_client.client
        # logger.info(f"CreateLimitOrder: quantity: {request.quantity}, price: {request.price}")
        try:
            res = await client.create_order(
                request.symbol,
                Side.BUY if request.buy_side else Side.SELL,
                order_type=OrderType.LIMIT,
                time_in_force=TimeInForce.GTC,
                quantity=request.quantity,
                quote_order_quantity=None,
                price=request.price,
                new_client_order_id=request.new_client_order_id,
                stop_price=None,
                iceberg_quantity=None,
                response_type=ResponseType.RESULT.value,
                receive_window=None,
                test=False)
        except errors.HTTPError as ex:
            logger.error(f"CreateLimitOrder for {open_client.name}:{request.symbol} exception: {ex.args}")
            context.set_details(f"{ex.args}")
            context.set_code(grpc.StatusCode.FAILED_PRECONDITION)
        except Exception as ex:
            logger.error(f"CreateLimitOrder for {open_client.name}:{request.symbol} exception: {ex}")
            context.set_details(f"{ex}")
            context.set_code(grpc.StatusCode.UNKNOWN)
        else:
            json_format.ParseDict(res, response)
        return response

    async def CancelOrder(self, request: api_pb2.CancelOrderRequest,
                          context: grpc.aio.ServicerContext) -> api_pb2.CancelOrderResponse:
        response = api_pb2.CancelOrderResponse()
        open_client = OpenClient.get_client(request.client_id)
        client = open_client.client
        try:
            res = await client.cancel_order(
                request.symbol,
                order_id=request.order_id,
                origin_client_order_id=None,
                new_client_order_id=None,
                receive_window=None)
        except asyncio.CancelledError:
            pass  # Task cancellation should not be logged as an error
        except errors.RateLimitReached as ex:
            Martin.rate_limit_reached_time = time.time()
            logger.warning(f"CancelOrder for {open_client.name}:{request.symbol} exception: {ex}")
            context.set_details(f"{ex}")
            context.set_code(grpc.StatusCode.RESOURCE_EXHAUSTED)
        except Exception as ex:
            logger.error(f"CancelOrder for {open_client.name}:{request.symbol} exception: {ex}")
            context.set_details(f"{ex}")
            context.set_code(grpc.StatusCode.UNKNOWN)
        else:
            json_format.ParseDict(res, response)
        return response

    async def StartStream(self, request: api_pb2.StartStreamRequest,
                          context: grpc.aio.ServicerContext) -> api_pb2.SimpleResponse:
        open_client = OpenClient.get_client(request.client_id)
        logger.debug(f"Start ws streams for {open_client.name}")
        response = api_pb2.SimpleResponse()
        _market_stream = None
        _market_stream_count = 0
        while _market_stream_count < request.market_stream_count:
            await asyncio.sleep(HEARTBEAT)
            _market_stream = open_client.client.events.registered_streams
            _market_stream_count = sum([len(_market_stream.get(k)) for k in _market_stream.keys()])
        # logger.debug(f"StartStream.events.registered_streams: {_market_stream}")
        # logger.debug(f"StartStream.events.handlers: {open_client.client.events.handlers}")
        asyncio.create_task(open_client.client.start_market_events_listener())
        asyncio.create_task(open_client.client.start_user_events_listener(endpoint=open_client.client.endpoint_ws))
        response.success = True
        return response

    async def StopStream(self, request: api_pb2.MarketRequest,
                         context: grpc.aio.ServicerContext) -> api_pb2.SimpleResponse:
        logger.info(f"StopStream")
        open_client = OpenClient.get_client(request.client_id)
        response = api_pb2.SimpleResponse()
        open_client.stop_streams_for_symbol = request.symbol
        [await _queue.put(None) for _queue in open_client.stream_queue]
        await open_client.client.stop_market_events_listener()
        await open_client.client.stop_user_events_listener()
        open_client.stream_queue = []
        gc.collect(generation=2)
        response.success = True
        return response


async def on_klines_update(_queue, event: events.KlineWrapper):
    # logger.info(f"on_klines_update.event: {event}")
    _event = weakref.ref(event)
    await _queue.put(_event())


async def on_order_update(_queue, event: events.OrderUpdateWrapper):
    # logger.debug(f"on_order_update.event: {event}")
    _event = weakref.ref(event)
    await _queue.put(_event())


async def on_funds_update(_queue, event: events.SymbolMiniTickerWrapper):
    # logger.info(f"on_funds_update.event: {event}")
    _event = weakref.ref(event)
    await _queue.put(_event())


async def on_ticker_update(_queue, event: events.SymbolMiniTickerWrapper):
    # logger.info(f"on_ticker_update.event: {event.event_type}, {event.event_time}")
    _event = weakref.ref(event)
    await _queue.put(_event())


async def on_order_book_update(_queue, event: events.PartialBookDepthWrapper):
    # logger.info(f"on_order_book_update.event: {event.last_update_id}")
    _event = weakref.ref(event)
    await _queue.put(_event())


async def serve() -> None:
    server = grpc.aio.server()
    api_pb2_grpc.add_MartinServicer_to_server(Martin(), server)
    listen_addr = '[::]:50051'
    server.add_insecure_port(listen_addr)
    logger.info(f"Starting server on {listen_addr}")
    await server.start()
    await server.wait_for_termination()


if __name__ == '__main__':
    FILE_LOG = f"{CONFIG.get('Path').get('log_path')}binance_srv.log"
    logger = logging.getLogger('exch_srv_logger')
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter(fmt="[%(asctime)s: %(levelname)s] %(message)s")
    #
    file_handler = logging.handlers.RotatingFileHandler(FILE_LOG, maxBytes=1000000, backupCount=10)
    file_handler.setFormatter(formatter)
    file_handler.setLevel(logging.DEBUG)
    logger.addHandler(file_handler)
    #
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(logging.INFO)
    logger.addHandler(stream_handler)
    #
    loop = asyncio.get_event_loop()
    loop.create_task(serve())
    loop.run_forever()
    loop.close()



#!/usr/bin/python3.8
# -*- coding: utf-8 -*-

from exchanges_wrapper import __version__

import aiohttp
import asyncio
import logging
import json
import hmac
import hashlib
import random

from ftx_parser import *

logger = logging.getLogger('exch_srv_logger')


class EventsDataStream:
    def __init__(self, client, endpoint, user_agent, exchange='binance'):
        self.client = client
        self.endpoint = endpoint
        if user_agent:
            self.user_agent = user_agent
        else:
            self.user_agent = f"exchange-wrapper, {__version__}"
        self.exchange = exchange
        self.web_socket = None
        self.try_count = 0

    async def start(self, *args, **kwargs):
        pass

    async def _handle_event(self, *args):
        pass

    async def _handle_messages(self, web_socket, symbol=None, ch_type=None):
        # logger.debug(f"_handle_messages.START: {web_socket}, symbol: {symbol}, ch_type: {ch_type}")
        self.try_count = 0
        msg_data_prev = {}
        while True:
            msg = await web_socket.receive()
            # print(f"_handle_messages.msg: {web_socket}, {msg}")
            if msg.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSE):
                logger.error(
                    "Trying to receive something while the websocket is closed! Trying to reconnect."
                )
                asyncio.ensure_future(self.start())
                break
            elif msg.type is aiohttp.WSMsgType.ERROR:
                logger.error(
                    f"Something went wrong with the websocket, reconnecting..."
                )
                asyncio.ensure_future(self.start())
                break
            elif msg.type == aiohttp.WSMsgType.CLOSING:
                logger.info(f"_handle_messages loop is stopped for {symbol} {ch_type}")
                break
            msg_data = json.loads(msg.data)
            # print(f"_handle_messages.msg_data: {msg_data}")
            if self.exchange == 'binance':
                # print(f"binance, channel: {ch_type}")
                await self._handle_event(msg_data)
            elif self.exchange == 'ftx' and msg_data.get('type') == 'update':
                # print(f"ftx, channel: {ch_type}")
                if msg_data.get('channel') not in ('fills', 'orders'):
                    _msg_data = eval(repr(msg_data))
                    if ftx_stream_compare(_msg_data, msg_data_prev, ch_type):
                        # print(f"_handle_messages.msg_data_LAST: {msg_data}")
                        msg_data_binance = ftx_stream_convert(msg_data, symbol, ch_type)
                        # print(f"_handle_messages.msg_data_binance: {msg_data_binance}")
                        await self._handle_event(msg_data_binance)
                    msg_data_prev = eval(repr(msg_data))
                else:
                    # print(f"_handle_messages.msg_data: {msg_data}")
                    msg_data_binance = ftx_stream_convert(msg_data)
                    # print(f"_handle_messages.msg_data_binance: {msg_data_binance}")
                    if msg_data_binance:
                        await self._handle_event(msg_data_binance)


class MarketEventsDataStream(EventsDataStream):
    def __init__(self, client, endpoint, user_agent, exchange='binance', channel=None):
        super().__init__(client, endpoint, user_agent, exchange)
        self.channel = channel

    async def stop(self):
        """
        Stop market data stream
        """
        logger.debug('MarketEventsDataStream.stop()')
        if self.web_socket:
            await self.web_socket.close()
        # print(f"stop.web_socket: {self.web_socket}, _state: {self.web_socket.closed}")

    async def start(self):
        registered_streams = self.client.events.registered_streams.get(self.exchange)
        logger.debug(f"MarketEventsDataStream.start().registered_streams: {registered_streams}")
        try:
            async with aiohttp.ClientSession() as session:
                if self.exchange == 'binance':
                    combined_streams = "/".join(registered_streams)
                    # print(f"start.combined_streams: {self.endpoint}/stream?streams={combined_streams}")
                    if self.client.proxy:
                        self.web_socket = await session.ws_connect(
                            f"{self.endpoint}/stream?streams={combined_streams}",
                            proxy=self.client.proxy,
                        )
                    else:
                        self.web_socket = await session.ws_connect(
                            f"{self.endpoint}/stream?streams={combined_streams}"
                        )
                    await self._handle_messages(self.web_socket)
                elif self.exchange == 'ftx':
                    if self.client.proxy:
                        self.web_socket = await session.ws_connect(
                            self.endpoint, heartbeat=15, proxy=self.client.proxy,)
                    else:
                        self.web_socket = await session.ws_connect(self.endpoint, heartbeat=15)
                    symbol = self.channel.split('@')[0]
                    ch_type = self.channel.split('@')[1]
                    request = {'op': 'subscribe', 'channel': 'ticker', 'market': symbol}
                    await self.web_socket.send_json(request)
                    await self._handle_messages(self.web_socket, symbol=symbol, ch_type=ch_type)
        except aiohttp.ClientOSError as ex:
            logger.error(f"MarketEventsDataStream.start(): try_count: {self.try_count}: {ex}")
            if self.try_count > 100:
                return
            self.try_count += 1
            await asyncio.sleep(random.uniform(0.2, 0.5) * self.try_count)
            asyncio.ensure_future(self.start())

    async def _handle_event(self, content):
        stream_name = None
        if "stream" in content:
            stream_name = content["stream"]
            content = content["data"]
        if isinstance(content, list):
            print('list')
            for event_content in content:
                event_content["stream"] = stream_name
                await self.client.events.wrap_event(event_content).fire()
        else:
            content["stream"] = stream_name
            await self.client.events.wrap_event(content).fire()


class FtxPrivateEventsDataStream(EventsDataStream):
    def __init__(self, client, endpoint, user_agent, exchange, sub_account=None):
        super().__init__(client, endpoint, user_agent, exchange)
        self.sub_account = sub_account

    async def stop(self):
        """
        Stop data stream
        """
        logger.info('FtxPrivateEventsDataStream.stop()')
        if self.web_socket:
            await self.web_socket.close()

    async def start(self):
        logger.info(f"FtxPrivateEventsDataStream.start(): exchange: {self.exchange}, endpoint: {self.endpoint}")
        async with aiohttp.ClientSession() as session:
            if self.client.proxy:
                self.web_socket = await session.ws_connect(
                    self.endpoint, heartbeat=15, proxy=self.client.proxy)
            else:
                self.web_socket = await session.ws_connect(self.endpoint, heartbeat=15)
            ts = int(time.time() * 1000)
            data = f"{ts}websocket_login"
            signature = hmac.new(self.client.api_secret.encode("utf-8"),
                                 data.encode("utf-8"),
                                 hashlib.sha256).hexdigest()
            request = {
                "op": "login",
                "args": {
                     "key": self.client.api_key,
                     "sign": signature,
                     "time": ts
                 }
            }
            if self.sub_account:
                request['args']['subaccount'] = self.sub_account
            await self.web_socket.send_json(request)
            request = {'op': 'subscribe', 'channel': 'fills'}
            await self.web_socket.send_json(request)
            request = {'op': 'subscribe', 'channel': 'orders'}
            await self.web_socket.send_json(request)
            await self._handle_messages(self.web_socket)

    async def _handle_event(self, content):
        # logger.debug(f"FtxPrivateEventsDataStream._handle_event.content: {content}")
        event = self.client.events.wrap_event(content)
        await event.fire()


class UserEventsDataStream(EventsDataStream):
    def __init__(self, client, endpoint, user_agent):
        super().__init__(client, endpoint, user_agent)

    async def _heartbeat(
        self, listen_key, interval=60 * 30
    ):  # 30 minutes is recommended according to
        # https://github.com/binance-exchange/binance-official-api-docs/blob/master/user-data-stream.md#pingkeep-alive-a-listenkey
        while True:
            await asyncio.sleep(interval)
            await self.client.keep_alive_listen_key(listen_key)

    async def stop(self):
        """
        Stop user data stream
        """
        # logger.info(f"UserEventsDataStream.stop: {self.client}")
        if self.web_socket:
            await self.web_socket.close()

    async def start(self):
        # logger.info(f"UserEventsDataStream.start: {self.client}")
        async with aiohttp.ClientSession() as session:
            listen_key = (await self.client.create_listen_key())["listenKey"]
            if self.client.proxy:
                self.web_socket = await session.ws_connect(
                    f"{self.endpoint}/ws/{listen_key}"
                )
            else:
                self.web_socket = await session.ws_connect(
                    f"{self.endpoint}/ws/{listen_key}", proxy=self.client.proxy
                )
            asyncio.ensure_future(self._heartbeat(listen_key))
            await self._handle_messages(self.web_socket)

    async def _handle_event(self, content):
        # print(f"user _handle_event.content: {content}")
        event = self.client.events.wrap_event(content)
        await event.fire()
